/*
 * Copyright 2015-2017 NetApp, Inc. All rights reserved.
 */

/*****************************************************************************
 *                             STOP RIGHT THERE!                             *
 *****************************************************************************
 *                                                                           *
 * HardwareConfigComponentMonitor and sfhwcheck are both involved in         *
 * hardware configuration checking and validation! If you are wanting to     *
 * create new cluster faults when hardware doesn't match expected values,    *
 * HardwareConfigComponentMonitor is the correct place to do so. However, if *
 * you are looking to do a one-time verification of hardware and potentially *
 * bail out of RTFI on a node, you should use sfhwcheck.                     *
 *                                                                           *
 * Please take some time to think over which file is more correct for your   *
 * desired checks, and implement them there.                                 *
 *****************************************************************************/


// <shared/common.h> MUST be first then shared/serviceshared
#include <shared/common.h>
#include <shared/AlgoString.h>
#include <shared/JsonUtil.h>
#include <shared/KeyDefine.h>
#include <shared/LoopUntil.h>
#include <shared/Shell.h>
#include <shared/Ioctl.h>
#include <shared/ProcFileSystem.h>

// Company

// System

// Module
#include "HardwareConfigComponentMonitor.h"

DEFINE_string(biosblacklist, "", "Comma separated list of settings to ignore when checking bios");

namespace COMPANY_NAMESPACE
{

const Path HardwareConfigComponentMonitor::kMptSasJsonPath = "/sf/etc/hardware/lsi.json";
const Path HardwareConfigComponentMonitor::kNicInfoPath = "/sf/etc/hardware/nic.json";
const Path HardwareConfigComponentMonitor::kSysNetworkPath = "/sys/class/net/";
const Path HardwareConfigComponentMonitor::kDellGenerationPath = "/opt/dell/srvadmin/lib/openmanage/IGNORE_GENERATION";
const Path HardwareConfigComponentMonitor::kSysPciPath = "/sys/bus/pci/devices/";
const Path HardwareConfigComponentMonitor::kBiosSettingsPath = "/var/log/sf-bios.info";

//Prevent flooding the system with cluster faults if something went awry.
const int  HardwareConfigComponentMonitor::kBadBiosSettingNum = 10;

const string ChassisAccessTriton::kSetupQuestionKey = "Setup Question";
const string ChassisAccessTriton::kBiosDefaultKey = "BIOS Default";
const string ChassisAccessTriton::kOptionsTypeKey = "Options";
const string ChassisAccessTriton::kValueTypeKey = "Value";

Mutex HardwareConfigComponentMonitor::mSerializedLock("HardwareConfigInternalCache", false);
Json::Value HardwareConfigComponentMonitor::mSerialized = Json::objectValue;

IMPL_KEYS(HardwareConfigComponentMonitor, HARDWARECONFIG_KEYS);

// BEGIN Originally located in driver/message/fusion/mptctl.h in the kernel.
#define MPT_IOCTL_VERSION_LENGTH (32)
struct mpt_ioctl_header {
    uint32_t ioc_number;
    uint32_t port_number;
    uint32_t max_data_size;
};
struct mpt_ioctl_pci_info {
    union {
        struct {
            uint32_t device:5;
            uint32_t function:3;
            uint32_t bus:24;
        } bits;
        uint32_t  word;
    } u;
    uint32_t segment_id;
};
struct mpt_ioctl_iocinfo {
    struct   mpt_ioctl_header hdr;
    uint32_t adapter_type;
    uint32_t port_number;
    uint32_t pci_id;
    uint32_t hw_rev;
    uint32_t subsystem_device;
    uint32_t subsystem_vendor;
    uint32_t rsvd0;
    uint32_t firmware_version;
    uint32_t bios_version;
    uint8_t  driver_version[MPT_IOCTL_VERSION_LENGTH];
    uint8_t  rsvd1;
    uint8_t  scsi_id;
    uint16_t rsvd2;
    struct mpt_ioctl_pci_info pci_information;
};
// END Originally located in driver/message/fusion/mptctl.h in the kernel.

// BEGIN Originally located in ethtool-copy.h in ethtool's source.
#define ETHTOOL_FWVERS_LEN      32
#define ETHTOOL_BUSINFO_LEN     32
#define ETHTOOL_EROMVERS_LEN    32

/**
 * struct ethtool_drvinfo - general driver and device information
 * @cmd: Command number = %ETHTOOL_GDRVINFO
 * @driver: Driver short name.  This should normally match the name
 *    in its bus driver structure (e.g. pci_driver::name).  Must
 *    not be an empty string.
 * @version: Driver version string; may be an empty string
 * @fw_version: Firmware version string; may be an empty string
 * @erom_version: Expansion ROM version string; may be an empty string
 * @bus_info: Device bus address.  This should match the dev_name()
 *    string for the underlying bus device, if there is one.  May be
 *    an empty string.
 * @n_priv_flags: Number of flags valid for %ETHTOOL_GPFLAGS and
 *    %ETHTOOL_SPFLAGS commands; also the number of strings in the
 *    %ETH_SS_PRIV_FLAGS set
 * @n_stats: Number of u64 statistics returned by the %ETHTOOL_GSTATS
 *    command; also the number of strings in the %ETH_SS_STATS set
 * @testinfo_len: Number of results returned by the %ETHTOOL_TEST
 *    command; also the number of strings in the %ETH_SS_TEST set
 * @eedump_len: Size of EEPROM accessible through the %ETHTOOL_GEEPROM
 *    and %ETHTOOL_SEEPROM commands, in bytes
 * @regdump_len: Size of register dump returned by the %ETHTOOL_GREGS
 *    command, in bytes
 *
 * Users can use the %ETHTOOL_GSSET_INFO command to get the number of
 * strings in any string set (from Linux 2.6.34).
 *
 * Drivers should set at most @driver, @version, @fw_version and
 * @bus_info in their get_drvinfo() implementation.  The ethtool
 * core fills in the other fields using other driver operations.
 */
struct ethtool_drvinfo {
    __u32    cmd;
    char     driver[32];
    char     version[32];
    char     fw_version[ETHTOOL_FWVERS_LEN];
    char     bus_info[ETHTOOL_BUSINFO_LEN];
    char     erom_version[ETHTOOL_EROMVERS_LEN];
    char     reserved2[12];
    __u32    n_priv_flags;
    __u32    n_stats;
    __u32    testinfo_len;
    __u32    eedump_len;
    __u32    regdump_len;
};
// END Originally located in ethtool-copy.h in ethtool's source.

void LogStanza(const vector<string>& stanza);

HardwareConfigComponentMonitor::HardwareConfigComponentMonitor(NodeID nodeID, boost::shared_ptr<DBClient> dbClient) :
    mNodeID(nodeID),
    mDBClient(dbClient),
    mDisableFaultChecks(false)
{ }

void HardwareConfigComponentMonitor::Check()
{
    LOG_DETAIL(Util) << "Checking hardware configuration.";
    bool allChecksPassed = true;

    const PlatformInfo& platform = PlatformInfo::GetInstance();
    if (platform.GetChassisType() == ChassisType::SFVIRT)
    {
        // Don't do any checks if we are a virtual node, just pretend it all passed
        GUARD(mSerializedLock);
        mSerialized = Json::objectValue;
        mSerialized[kPassedKey] = true;
        mSerialized[kHardwareConfigKey] = Json::objectValue;
        return;
    }

    Json::Value latestInfo = Json::objectValue;

    allChecksPassed = CheckCpu(platform, latestInfo) && allChecksPassed;
    allChecksPassed = CheckMemory(platform, latestInfo) && allChecksPassed;
    allChecksPassed = CheckSas(platform, latestInfo) && allChecksPassed;
    allChecksPassed = CheckChassis(platform, latestInfo) && allChecksPassed;
    allChecksPassed = CheckNetworking(platform, latestInfo) && allChecksPassed;

    /*
     * NOTE - The following lines should always be the last things to happen in this function!
     */
    LOG_DETAIL(Util) << "Checking hardware configuration completed," << LVAL(allChecksPassed);
    latestInfo[kPassedKey] = allChecksPassed;

    GUARD(mSerializedLock);
    mSerialized = latestInfo;
}

template<typename T> bool Compare(const string& a, const string& op, const T& b)
{
    if (op == "ANY")
    {
        return true;
    }

    TRY
    {
        if(op == ">=")
        {
            return FromJson<T>(a) >= b;
        }
        else if(op == "<=")
        {
            return FromJson<T>(a) <= b;
        }
        return FromJson<T>(a) == b;
    }
    CATCH(ex)
    {
        return false;
    }
}
template<> bool Compare<string>(const string& a, const string& op, const string& b)
{
    if (op == "ANY")
    {
        return true;
    }

    if(op == ">=")
    {
        return VersionCompareLte(b, a);
    }
    else if(op == "<=")
    {
        return VersionCompareLte(a, b);
    }
    else if (op == "!=")
    {
        return a != b;
    }
    return a == b;
}

template<typename T> bool HardwareConfigComponentMonitor::RunCheckInternal(
        const string& key,
        const string& actual,
        const string& sop,
        const T& expected,
        INOUT Json::Value& latestInfo,
        bool forcePass /* default: false */)
{
    string ac = boost::trim_copy(actual);
    string ex = boost::trim_copy(ToString(expected));
    // If forcePass == true, pass will always be true, regardless of what compare returns (since we short-circuit it).
    bool pass = forcePass || Compare(ac, sop, expected);
    latestInfo[kHardwareConfigKey][key] = CheckInternal(key, ac, sop, ex, pass);
    return pass;
}

Json::Value HardwareConfigComponentMonitor::CheckInternal(const string& key, const string& actual, const string& op, const string& expected, bool pass)
{
    LOG_DETAIL(Util) << "Checking" << LVALS(key, actual, op, expected);

    if (LIKELY(!mDisableFaultChecks))
    {
        // Create an entry for this key (if we haven't before)
        if (mFaults.count(key) == 0)
        {
            mFaults.emplace(key, NodeHardwareFault(mNodeID, mDBClient));
        }

        // Grab reference to existing Fault from our map
        NodeHardwareFault& fault = mFaults.at(key);

        // If it passed resolve any existing fault that may be present otherwise create a new fault.
        if (UNLIKELY(!pass))
        {
            string actualValue = ToString(actual);
            fault.Set(ClusterFaultSeverity::Warning,
                      ToStringFormat("%1% should be %2% %3% but is %4%", key, op, ToString(expected), actualValue.empty() ? "null" : actualValue));
        }
        else
        {
            fault.Resolve();
        }
    }

    Json::Value ret = Json::objectValue;
    ret[kPassedKey] = pass;
    ret[kActualKey] = actual;
    ret[kOpKey] = op;
    ret[kExpectedKey] = expected;
    return ret;
}

Json::Value HardwareConfigComponentMonitor::GetData()
{
    HardwareConfigComponentMonitor monitor(NodeID::None(), nullptr);
    monitor.mDisableFaultChecks = true;
    monitor.Check();

    GUARD(mSerializedLock);
    return mSerialized;
}

void HardwareConfigComponentMonitor::GetSerializedComponentInfo(INOUT Json::Value& serialized) const
{
    GUARD(mSerializedLock);
    JsonMemberInto(mSerialized, kHardwareConfigKey, serialized[kHardwareConfigKey]);
}

bool HardwareConfigComponentMonitor::CheckCpu(const PlatformInfo& platform, Json::Value& latestInfo)
{
    LOG_DETAIL(Util) << "Checking hardware:CPU configuration";
    bool allChecksPassed = true;
    int numCpuThreads = 0;
    ShellResult cpuCount = Shell("grep 'physical id' /proc/cpuinfo | sort -u | wc -l").Trim();
    allChecksPassed &= RunCheckInternal("NUM_CPU", cpuCount.stdout, "==", platform.GetNumCpu(), latestInfo);

    int counts[] = {0, 0, 0, 0};
    for (const string& line : Shell("dmidecode -t processor").Split())
    {
        vector<string> parts = Split(line, ":");
        if(parts.size() == 2)
        {
            // Trim out whitespace from the front and back of the strings.
            string key   = boost::trim_copy(parts.at(0));
            string value = boost::trim_copy(parts.at(1));

            if(key == "Version")
            {
                allChecksPassed &= RunCheckInternal(ToStringFormat("CPU_MODEL_%02d", counts[0]), value, "==", platform.GetCpuModel(), latestInfo);
                counts[0]++;
            }
            else if(key == "Core Count")
            {
                allChecksPassed &= RunCheckInternal(ToStringFormat("CPU_CORES_%02d", counts[1]), value, "==", platform.GetCpuCores(), latestInfo);
                counts[1]++;
            }
            else if(key == "Core Enabled")
            {
                allChecksPassed &= RunCheckInternal(ToStringFormat("CPU_CORES_ENABLED_%02d", counts[2]), value, "==", platform.GetCpuCoresEnabled(), latestInfo);
                counts[2]++;
            }
            else if(key == "Thread Count")
            {
                numCpuThreads = FromJson<int>(value);
                allChecksPassed &= RunCheckInternal(ToStringFormat("CPU_THREADS_%02d", counts[3]), value, "==", platform.GetCpuThreads(), latestInfo);
                counts[3]++;
            }
        }
    }

    ShellResult coreCount = Shell("grep -c '^processor' /proc/cpuinfo").Trim();
    allChecksPassed &= RunCheckInternal("CPU_THREADS_ENABLED", coreCount.stdout, "==", FromJson<int>(cpuCount.stdout) * numCpuThreads, latestInfo);

    return allChecksPassed;
}

bool HardwareConfigComponentMonitor::CheckMemory(const PlatformInfo& platform, INOUT Json::Value& latestInfo)
{
    LOG_DETAIL(Util) << "Checking hardware:memory configuration";
    bool allChecksPassed = true;

    ShellResult memorySpeeds = Shell("dmidecode --type memory | grep 'Size:\\\\|Configured Clock Speed:' | grep -v 'Unknown\\\\|No Module Installed'");
    uint64_t memoryMB =  0;
    int iteration = 0;
    for (const string& line : memorySpeeds.Split())
    {
        vector<string> parts = Split(line, ":");
        if(parts.size() == 2)
        {
            // Trim out whitespace from the front and back of the strings.
            string key   = boost::trim_copy(parts.at(0));
            string value = boost::trim_copy(parts.at(1));

            if(key == "Size")
            {
                vector<string> item = Split(value, " ");;
                if(item[1] == "MB")
                {
                    memoryMB += FromJson<uint64_t>(item[0]);
                }
                else if(item[1] == "GB")
                {
                    memoryMB += FromJson<uint64_t>(item[0]) * KiByte;
                }

            }
            else if(key == "Configured Clock Speed")
            {
                vector<string> item = Split(value, " ");
                allChecksPassed &= RunCheckInternal(ToStringFormat("MEMORY_MHZ_%02d", iteration / 2), item[0], ">=", platform.GetMemoryMhz(), latestInfo);
            }
            iteration++;
        }
    }
    allChecksPassed &= RunCheckInternal("MEMORY_GB", ToString(memoryMB / KiByte), ">=", platform.GetMemoryGB(), latestInfo);

    return allChecksPassed;
}

bool HardwareConfigComponentMonitor::CheckSas(const PlatformInfo& platform, INOUT Json::Value& latestInfo)
{
    bool allChecksPassed = true;
    LOG_DETAIL(Util) << "Checking hardware:SAS configuration";

    // Checking sas card values, but only if expect a sas controller
    if (platform.GetScsiBusExternalDriver() == "")
    {
        return allChecksPassed;
    }

    bool foundSasDevice = false;
    IProcFileSystem& procfs = ProcFileSystem::GetInstance();

    SasInfo sasInfo;
    for (const Path& path : FileSystem::ListDirectory(kSysPciPath))
    {
        string vendor       = procfs.Read(path.ToString() + "/vendor").GetFirstLineValue().substr(2);
        string device       = procfs.Read(path.ToString() + "/device").GetFirstLineValue().substr(2);
        string subSysVendor = procfs.Read(path.ToString() + "/subsystem_vendor").GetFirstLineValue().substr(2);
        string subSysDevice = procfs.Read(path.ToString() + "/subsystem_device").GetFirstLineValue().substr(2);

        TRY
        {
            sasInfo = platform.GetSasInfo(vendor, device, subSysVendor, subSysDevice);
            foundSasDevice = true;
            // Assuming this worked, we've found our device
            break;
        }
        CATCH(ex)
        {
            // This isn't the device you're looking for...
        }
    }

    if (!foundSasDevice || platform.GetScsiBusExternalDriver() != "mpt3sas")
    {
        static const string kUnknownVal = "Unknown";
        LOG_EXPERR(Util) << "No Recognized SAS Device or Unrecognized Driver"
            << LVALBOOL(foundSasDevice) << LVAL(platform.GetScsiBusExternalDriver());
        allChecksPassed &= RunCheckInternal("MPTSAS_FIRMWARE_VERSION", kUnknownVal, "!=", kUnknownVal, latestInfo);
        allChecksPassed &= RunCheckInternal("MPTSAS_BIOS_VERSION", kUnknownVal, "!=", kUnknownVal, latestInfo);
        return allChecksPassed;
    }

    mpt_ioctl_iocinfo mptInfo;
    memset(&mptInfo, 0, sizeof(mptInfo));
    bool mptRet = Ioctl::TryCall(sasInfo.GetDevPath(), _IOWR(0x4c, 17, struct mpt_ioctl_iocinfo), &mptInfo);
    LOG_DETAIL(Util) << "MPT Ioctl " << LVALBOOL(mptRet);

    string mptsasFirmwareVersion = "";
    string mptsasBiosVersion = "";
    if (mptRet)
    {
        mptsasFirmwareVersion = ToStringFormat("%02d.%02d.%02d.%02d",
           (mptInfo.firmware_version >> (8 * 3)) & 0xff,
           (mptInfo.firmware_version >> (8 * 2)) & 0xff,
           (mptInfo.firmware_version >> (8 * 1)) & 0xff,
           (mptInfo.firmware_version >> (8 * 0)) & 0xff);
        mptsasBiosVersion = ToStringFormat("%02d.%02d.%02d.%02d",
           (mptInfo.bios_version >> (8 * 3)) & 0xff,
           (mptInfo.bios_version >> (8 * 2)) & 0xff,
           (mptInfo.bios_version >> (8 * 1)) & 0xff,
           (mptInfo.bios_version >> (8 * 0)) & 0xff);
    }

    SoftwareVersion expectedFirmware;
    if(platform.GetChassisType() == ChassisType::R630 && sasInfo.GetR630FirmwareVersion().IsNotZero())
    {
        expectedFirmware = sasInfo.GetR630FirmwareVersion();
    }
    else
    {
        expectedFirmware = sasInfo.GetFirmwareVersion();
    }
    allChecksPassed &= RunCheckInternal("MPTSAS_FIRMWARE_VERSION", mptsasFirmwareVersion, "==", expectedFirmware, latestInfo);

    // MPTSAS_BIOS_VERSION will _always_ pass and never cause a fault as per FOGBUGZ-21009.
    // FOGBUGZ-21031 - Remove and verify removal of MPTSAS_BIOS or make a fault.
    // FOGBUGZ-23027 - Add "ANY" operator support so people don't get confused...
    allChecksPassed &= RunCheckInternal("MPTSAS_BIOS_VERSION", mptsasBiosVersion, "ANY", sasInfo.GetBiosVersion(), latestInfo, true);

    return allChecksPassed;
}

bool HardwareConfigComponentMonitor::CheckChassis(const PlatformInfo& platform, INOUT Json::Value& latestInfo)
{
    LOG_DETAIL(Util) << "Checking hardware:BIOS and chassis configuration";
    bool allChecksPassed = true;

    ShellResult biosInfo = Shell("dmidecode --type bios");
    for (const string& line : biosInfo.Split())
    {
        vector<string> parts = Split(line, ":");
        if(parts.size() == 2)
        {
            // Trim out whitespace from the front and back of the strings.
            string key   = boost::trim_copy(parts.at(0));
            string value = boost::trim_copy(parts.at(1));

            if(key == "Vendor")
            {
                vector<string> vendorArray = platform.GetBiosVendor();
                string expectedVendor = ToString(vendorArray); // reported when we don't find a match
                for(const string& vendor : vendorArray)
                {
                    string vendorString = boost::trim_copy(vendor);
                    if(vendorString == value)
                    {
                        // when we find a match in the array, overwrite the expectedVendor array
                        // with the actual match so our check passes below
                        expectedVendor = vendorString;
                        break;
                    }
                }
                allChecksPassed &= RunCheckInternal("BIOS_VENDOR", value, "==", expectedVendor, latestInfo);
            }
            else if(key == "Version")
            {
                allChecksPassed &= RunCheckInternal("BIOS_VERSION", value, ">=", platform.GetBiosVersion(), latestInfo);
            }
            else if(key == "BIOS Revision")
            {
                allChecksPassed &= RunCheckInternal("BIOS_REVISION", value, ">=", platform.GetBiosRevision(), latestInfo);
            }
        }
    }

    allChecksPassed &= CheckBiosSettings(platform, latestInfo);

    if (platform.GetChassisType() == ChassisType::R620 || platform.GetChassisType() == ChassisType::R630)
    {
        ShellResult idracInfo = TryShell("ipmi-oem dell get-system-info idrac-info");

        for (const string& line : idracInfo.Split())
        {
            vector<string> parts = Split(line, ":");
            if(parts.size() == 2)
            {
                // Trim out whitespace from the front and back of the strings.
                string key   = boost::trim_copy(parts.at(0));
                string value = boost::trim_copy(parts.at(1));

                if (key == "iDRAC Firmware Version")
                {
                    vector<string> versionParts = Split(value, " ");
                    if (versionParts.size() >= 2)
                    {
                        allChecksPassed &= RunCheckInternal("IDRAC_VERSION", versionParts[0], ">=", platform.GetIdracVersion(), latestInfo);
                    }
                    else
                    {
                        allChecksPassed &= RunCheckInternal("IDRAC_VERSION", value, ">=", platform.GetIdracVersion(), latestInfo);
                    }
                }
            }
        }
    }

    return allChecksPassed;
}

boost::shared_ptr<ChassisAccess> HardwareConfigComponentMonitor::ChassisAccessFactory(const PlatformInfo& platform)
{
    std::map<ChassisType, boost::shared_ptr<ChassisAccess> >::iterator iter = mChassisAccess.find(platform.GetChassisType());

    if (iter != mChassisAccess.end())
    {
        return iter->second;
    }

    boost::shared_ptr<ChassisAccess> chassisAccess;
    TRY
    {
        switch (platform.GetChassisType())
        {
        case ChassisType::QS52B:
            chassisAccess = boost::make_shared<ChassisAccessTriton>();
            break;
        case ChassisType::SFVIRT:
        case ChassisType::R620:
        case ChassisType::R630:
        case ChassisType::C220M4:
        case ChassisType::SMBigTwin:
        case ChassisType::Unknown:
        default:
            chassisAccess = boost::make_shared<ChassisAccess>(platform.GetChassisType());
        }

        mChassisAccess[chassisAccess->GetChassisType()] = chassisAccess;
    }
    CATCH(ex)
    {
        LOG_UNHERR(Util) << "Received an exception:" << LVAL(ex);
    }

    return chassisAccess;
}

bool HardwareConfigComponentMonitor::CheckBiosSettings(const PlatformInfo& platform, INOUT Json::Value& latestInfo)
{
    boost::shared_ptr<ChassisAccess> chassisAccess = ChassisAccessFactory(platform);

    if (!chassisAccess || chassisAccess->Checked())
    {
        return true;
    }

    int numBadAllowed = kBadBiosSettingNum;

    vector<BiosSetting> biosSettings;
    bool settingsCorrect = chassisAccess->GetBiosSettings(biosSettings);

    if (!RunCheckInternal(ToStringFormat("Parse failure in %s", kBiosSettingsPath.ToString()),
                                        settingsCorrect ? "false" : "true", "==", string("false"), latestInfo))
    {
        return false;
    }

    for (vector<BiosSetting>::iterator iter = biosSettings.begin(); iter != biosSettings.end() && numBadAllowed; ++iter)
    {
        BiosSetting& biosSetting = *iter;
        bool correct = RunCheckInternal(ToStringFormat("Incorrect bios setting \"%s\"", biosSetting.GetName()), biosSetting.GetActual(), "==", biosSetting.GetExpected(), latestInfo);
        settingsCorrect &= correct;
        if (!correct)
        {
            --numBadAllowed;
        }
    }

    return settingsCorrect;
}

bool HardwareConfigComponentMonitor::CheckNetworking(const PlatformInfo& platform, INOUT Json::Value& latestInfo)
{
    LOG_DETAIL(Util) << "Checking hardware:networking configuration";
    bool allChecksPassed = true;

    // FOGBUGZ-23016 - For Cisco NIC, we do not have logic to verify Cisco NIC firmware version and will need to be addressed
    bool isNodeTypeSF9608 = platform.GetNodeType() == NodeType::SF9608;

    int sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock < 0)
    {
        sock = socket(AF_NETLINK, SOCK_RAW, NETLINK_GENERIC);
    }
    FileHandle fd(sock);

    ifreq ifr;
    ethtool_drvinfo ethDriverInfo;
    set<Path> networkDevices = FileSystem::ListDirectory(kSysNetworkPath);
    for (const Path& netdev : networkDevices)
    {
        if (FileSystem::IsSymLink(netdev) && FileSystem::Exists(netdev + "/device"))
        {
            string device = netdev.FileName().ToString();

            memset(&ifr, 0, sizeof(ifr));
            strncpy(ifr.ifr_name, device.c_str(), min(IFNAMSIZ, static_cast<int>(1 + device.size()))); // +1 for '\0' character.
            ethDriverInfo.cmd = 0x3; // ETHTOOL_GDRVINFO
            ifr.ifr_data = reinterpret_cast<char*>(&ethDriverInfo);

            NicInfo nicInfo = platform.GetNicInfoFromSystem(device);
            string expectedFirmware = nicInfo.GetVersion();

            string actualDriver = "";
            string actualFirmware = "";
            bool ret = Ioctl::TryCall(fd, SIOCETHTOOL, &ifr);

            if (ret) // success
            {
                actualDriver = GetBoundedLengthString(ethDriverInfo.driver, (int)sizeof(ethDriverInfo.driver));
                actualFirmware = GetBoundedLengthString(ethDriverInfo.fw_version, (int)sizeof(ethDriverInfo.fw_version));
            }

            boost::to_upper(device);

            if (expectedFirmware != actualFirmware)
            {
                for (const string& firmwareVersion : nicInfo.GetLegacyVersions())
                {
                    if (firmwareVersion == actualFirmware)
                    {
                        expectedFirmware = firmwareVersion;
                        break;
                    }
                }
            }
            allChecksPassed &= RunCheckInternal(ToStringFormat("NETWORK_FIRMWARE_VERSION_%s", device), actualFirmware, "==", expectedFirmware, latestInfo, isNodeTypeSF9608);

            vector<string> driverArray = platform.GetNetworkDriver();
            string expectedDriver = ToString(driverArray);
            for (const string& driver : driverArray)
            {
                string driverString = boost::trim_copy(driver);
                if (driverString == actualDriver)
                {
                    expectedDriver = driverString;
                    break;
                }
            }

            allChecksPassed &= RunCheckInternal(ToStringFormat("NETWORK_DRIVER_%s", device), actualDriver, "==", expectedDriver, latestInfo, isNodeTypeSF9608);
        }
    }

    if (platform.GetFibreChannelPorts().size() != 0)
    {
        // These checks are all taken care of by the Fibre Channel monitor,
        // however we need to be backwards compatible with the API so we still
        // need the keys to show up in what we respond with. For that reason,
        // we force them to pass the check and give them no value.
        allChecksPassed &= RunCheckInternal<string>("NUM_FIBRE_CHANNEL_PORTS", "", "==", "", latestInfo, true);
        allChecksPassed &= RunCheckInternal<string>("FIBRE_CHANNEL_INCONSISTENCY", "", "==", "", latestInfo, true);
        allChecksPassed &= RunCheckInternal<string>("FIBRE_CHANNEL_MODEL", "", "==", "", latestInfo, true);
        allChecksPassed &= RunCheckInternal<string>("FIBRE_CHANNEL_FIRMWARE_REVISION", "", "==", "", latestInfo, true);
    }

    return allChecksPassed;
}


bool ChassisAccess::GetBiosSettings(OUT vector<BiosSetting>& biosSettings)
{
    biosSettings.clear();
    SetChecked(true);
    return true;
}

vector<string> ChassisAccess::GetBiosBlacklist()
{
    vector<string> blacklist = Split(FLAGS_biosblacklist, ",");
    vector<string> platformBiosBlacklist = PlatformInfo::GetInstance().GetBiosBlacklist();
    blacklist.insert(end(blacklist), begin(platformBiosBlacklist), end(platformBiosBlacklist));
    return blacklist;
}

// This command will report settings changed from the default (and settings that the BIOS has dynamically modified),
// but for extensibility the CheckBiosSettings method was designed to check the expected value against the actual value.
// As such, we parse the output of the command below and grab both the expected and actual values and feed them to the algorithm.
string ChassisAccessTriton::mBiosSettingsCmd = "/sf/packages/ami-tools-solidfire-1.0/bin/scelnx_64 /o /s %1% /v /q /hb /ndef";

ChassisAccessTriton::ChassisAccessTriton() :
    ChassisAccess(ChassisType::QS52B)
{}

bool ChassisAccessTriton::GetBiosSettings(OUT vector<BiosSetting>& biosSettings)
{
    FileSystem::Remove(HardwareConfigComponentMonitor::kBiosSettingsPath);
    biosSettings.clear();

    ShellResult shellResult = TryShell(mBiosSettingsCmd.c_str(), HardwareConfigComponentMonitor::kBiosSettingsPath.ToString());
    if (!shellResult)
    {
        LOG_APP(Util) << "Command failed \"" << mBiosSettingsCmd << "\"" << HardwareConfigComponentMonitor::kBiosSettingsPath.ToString() << LVAL(shellResult);
        SetChecked(false);
        return false;
    }

    /*  Examples of what to expect when parsing the output of the command above.

        Setup Question  = Package C State
        Token   =1790   // Do NOT change this line
        Offset  =2C
        Width   =01
        BIOS Default =[00]C0/C1 state
        MFG Default =[00]C0/C1 state
        Options =[00]C0/C1 state        // Move "*" to the desired Option
                 [01]C2 state
                *[02]C6(non Retention) state
                [03]C6(Retention) state
                [07]No Limit
                [FF]Auto

        Setup Question  = VccSA/VccIO Disable
        Token   =1791   // Do NOT change this line
        Offset  =FA
        Width   =01
        BIOS Default =[00]Disable
        MFG Default =[00]Disable
        Options =*[00]Disable   // Move "*" to the desired Option
                 [01]Enable

        Setup Question  = C2C3TT
        Token   =1792   // Do NOT change this line
        Offset  =29
        Width   =01
        BIOS Default =<0>
        MFG Default =<0>
        Value   =<0>

        Setup Question  = Reset Platform on Memory Map Change
        Token   =0D     // Do NOT change this line
        Offset  =CD
        Width   =01
        BIOS Default =0
        MFG Default =0
        Value   =0      // Enabled = 1, Disabled = 0

        Setup Question  = IPv6 Router1 IP Address
        Token   =80C    // Do NOT change this line
        Offset  =94
        Width   =4E
        Value   =""

    */

    vector<string> fileContents = BasicFile::ReadContentsSplit(HardwareConfigComponentMonitor::kBiosSettingsPath, "\n", false);
    vector<string>::const_iterator curIter = fileContents.begin();
    vector<string>::const_iterator endIter = fileContents.end();
    vector<string> blacklist = GetBiosBlacklist();

    LOG_APP(Util) << "Parsing started, num lines: " << fileContents.size();

    do
    {
        vector<string> curStanza = GetNextStanza(curIter, endIter);

        string setupQuestionValue = GetSetupQuestionValue(curStanza);
        if (setupQuestionValue.empty())
        {
            break;
        }

        if (std::find(blacklist.begin(), blacklist.end(), setupQuestionValue) != blacklist.end())
        {
            LOG_DETAIL(Util) << "Ignoring" << LVAL(blacklist) << LVALBR(setupQuestionValue);
            continue;
        }

        string biosDefaultValue = GetBiosDefaultValue(curStanza);
        if (biosDefaultValue.empty())
        {
            //Some stanzas do not contain a default value.
            //Ignore these entries.
            continue;
        }

        vector<string> biosCurrentValues = GetBiosCurrentValues(curStanza);
        string biosCurrentValue;
        if (biosCurrentValues.empty())
        {
            break;
        }
        else
        {
            //Some stanzas have multiple values specified.
            //Determine if any match the default value.
            auto iter = find(biosCurrentValues.begin(), biosCurrentValues.end(), biosDefaultValue);
            if (iter != biosCurrentValues.end())
            {
                biosCurrentValue = *iter;
            }
            else
            {
                biosCurrentValue = biosCurrentValues.at(0);
            }
        }

        biosSettings.push_back(BiosSetting(setupQuestionValue, biosDefaultValue, biosCurrentValue));

    } while (curIter != endIter);

    //We consider parsing a success if the file had contents and we iterated over everything.
    SetChecked(fileContents.size() && curIter == endIter);

    LOG_APP(Util) << "Parsing complete, success: " << Checked() << LVAL(biosSettings.size());

    return Checked();
}

vector<string> ChassisAccessTriton::GetNextStanza(INOUT vector<string>::const_iterator& curIter, const vector<string>::const_iterator endIter)
{
    vector<string> stanza;
    bool start = false;
    for (; curIter != endIter; ++curIter)
    {
        const string& line = *curIter;

        if (boost::starts_with(line, kSetupQuestionKey))
        {
            start = true;
        }
        else if (start && line == "")
        {
            break;
        }

        if (start)
        {
            stanza.push_back(boost::trim_copy(line.substr(0, line.find("\n"))));
        }
    }

    //LogStanza(stanza);
    return stanza;
}

void LogStanza(const vector<string>& stanza)
{
    LOG_APP(Util) << "---- Stanza Begin ----";
    for (auto line : stanza)
    {
        LOG_APP(Util) << "\"" << line << "\"";
    }
    LOG_APP(Util) << "---- Stanza End ----";
}

string ChassisAccessTriton::GetValueForKey(const string& key, const vector<string>& stanza)
{
    string value;
    find_if(begin(stanza), end(stanza),
                        [&] (const string& input)
                        {
                            string line = boost::trim_copy(input);
                            if (boost::starts_with(line, key))
                            {
                                value = line;
                                return true;
                            }
                            return false;
                        });

    if (value.empty())
    {
        return value;
    }

    vector<string> parts = Split(value, "=");
    if (parts.size() < 2)
    {
        // Error parsing file.
        LOG_APP(Util) << "Malformed stanza - does not contain '='" << LVAL(stanza);
        return string();
    }

    value = parts.at(1);

    //Split at comment (if it exists).
    return boost::trim_copy(value.substr(0, value.find("//")));
}

string ChassisAccessTriton::GetSetupQuestionValue(const vector<string>& stanza)
{
    return GetValueForKey(kSetupQuestionKey, stanza);
}

string ChassisAccessTriton::GetBiosDefaultValue(const vector<string>& stanza)
{
    return GetValueForKey(kBiosDefaultKey, stanza);
}

vector<string> ChassisAccessTriton::GetBiosCurrentValues(const vector<string>& stanza)
{
    //Check to see if this stanza is a "Value" type stanza first.
    string curValue = GetValueForKey(kValueTypeKey, stanza);
    vector<string> values;
    if (!curValue.empty())
    {
        values.push_back(curValue);
        return values;
    }

    /*  This could be an "Options" type stanza.
        Examples: (All spaces around lines have already been removed)

        Setup Question  = Additional Post Codes
        Token   =18     // Do NOT change this line
        Offset  =1C3
        Width   =01
        BIOS Default =[00]Disable
        MFG Default =[00]Disable
        Options =*[00]Disable   // Move "*" to the desired Option
                    [01]Enable


        Setup Question  = Advanced Debug Function
        Token   =17     // Do NOT change this line
        Offset  =1C2
        Width   =01
        BIOS Default =[02]Auto
        MFG Default =[02]Auto
        Options =*[02]Auto      // Move "*" to the desired Option
                    [00]Disable
                    [01]Enable

        Setup Question  = PPR Type
        Token   =162E   // Do NOT change this line
        Offset  =11C
        Width   =01
        BIOS Default =[00]PPR Disabled
        MFG Default =[00]PPR Disabled
        Options =*[00]Auto      // Move "*" to the desired Option
                    [01]Hard PPR
                    [02]Soft PPR
                    *[00]PPR Disabled
    */

    bool start = false;
    for (auto line : stanza)
    {
        curValue = line;
        if (boost::starts_with(curValue, kOptionsTypeKey))
        {
            start = true;

            vector<string> parts = Split(curValue, "=");
            if (parts.size() < 2)
            {
                // Error parsing file.
                LOG_APP(Util) << "Malformed stanza - does not contain '=' in Options line" << LVAL(stanza);
                return vector<string>();
            }

            curValue = parts.at(1);
        }

        if (start)
        {
            if (boost::starts_with(curValue, "*["))
            {
                //Get rid of the "*"
                curValue = curValue.substr(1);

                //Split at comment (if it exists).
                values.push_back(boost::trim_copy(curValue.substr(0, curValue.find("//"))));
            }
        }
    }

    return values;
}

} // end COMPANY_NAMESPACE

